#
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
# default deceit Perl library
#
# Giuseppe Vacanti, XMM SOC, 1998
#
#

BEGIN
  {
    use deceit;
    deceit::addPerlFunctionToLookupTable("rows");
    deceit::addPerlFunctionToLookupTable("fillArray");
    deceit::addPerlFunctionToLookupTable("fillArrayFromFits");
    deceit::addPerlFunctionToLookupTable("fillArrayFromTuple");
    deceit::addPerlFunctionToLookupTable("fillConstArray");
    deceit::addPerlFunctionToLookupTable("fillConstColumn");
    deceit::addPerlFunctionToLookupTable("fillColumn");
    deceit::addPerlFunctionToLookupTable("fillAttribute");
    deceit::addPerlFunctionToLookupTable("callUserFunction");
  }

require 5.004;
use LineParser;
use Env;
use FileHandle;
use Carp;
use DAL;
use DAL::utils qw(setStringCell);
use strict;
$| = 1;

#
# These are utility functions that brutally cast the pointer passed by
# the deceit interpreter into a pointer of the appropriate type. 
#
sub makeColumnPtr
{
    my $ptr = shift or croak("No pointer specified.\n");
    bless \$ptr, 'ColumnPtr';
}

sub makeArrayPtr
{
    my $ptr = shift or croak("No pointer specified.\n");
    bless \$ptr, 'ArrayPtr';
}

sub makeDatasetPtr
{
    my $ptr = shift or croak("No pointer specified.\n");
    bless \$ptr, 'DatasetPtr';
}
sub makeAttributePtr {
    my $ptr = shift or croak("No pointer specified.\n");
    bless \$ptr, 'AttributePtr';
}


sub min
{
    my($a,$b) = @_;    
    $a <= $b ? $a : $b;
}
sub max
{
    my($a,$b) = @_;    
    $a < $b ? $b : $a;
}


#
# rows(<filename>)
#
#  returns the number of rows in a given file (normally ASCII)
#
#   + blank lines do not count (match /^\s*$/)
#   + comment lines do not count (match /^\s*[#!]/)
#
sub rows
{
    my $fname = shift || do {
	carp("No file name given. Number of rows set to 0.\n");
	return 0;
    };
    my $fh = FileHandle->new($fname, "r") || do {
	carp("$fname does not exist. Number of rows set to 0.\n");
	return 0;
    };
    my $nrows = 0;
    while(<$fh>) {
	next if /^\s*[#!]/; # skip comment lines
	next if /^\s*$/; # skip blank lines
       $nrows++;
    }
    undef $fh;
    print "There are $nrows rows in $fname\n" if($ENV{"VERBOSE"});
    return $nrows;
}
    
#
# fillColumn(<ptr>, <filename>, <colnum> [, <sep> [, <padding>]]) 
#
#  fills a DAL column with input from file
#
#   + <ptr>: pointer to the column to be filled in (variable col from deceit)
#   + <filename>: file containing the ASCII data to be read into the
#                 column
#   + <colnum>: column number (first column = 0) in the ASCII file
#               from which the data should be read
#   + <sep>: column separator. This is optional, defaults to
#            space. This is mostly useful when using vector columns.
#   + <padding>: if true then numerical vector columns will be padded with 
#                trailing 0s if not enough elements are found.
#
sub fillColumn
  {
    if (@_ < 3 or @_ > 5) {
      croak("Usage: fillColumn(col, \$fname, \$colnum [,\$sep [,\$padding]])\n");
    }
    my ($ptr, $fname, $colnum, $sep, $padding) = @_;
    $sep = ' ' unless defined $sep;
    LineParser->setSeparator($sep);
    # this is a hack: take the address of the column (passed from
    # deceit), cast it into a pointer to a column, then stuff it into a
    # column object.
    my $p = makeColumnPtr($ptr);
    my $col = DAL::column->_new();
    $col->ptr($p); 
    printf STDERR "\nFilling column %s from %s (column %d, separator is >$sep<) -->",
    $col->name, $fname, $colnum, $sep if($ENV{"VERBOSE"});
    my $fh = FileHandle->new();
    $fh->open($fname, "r") || do {
      carp("$fname does not exist. The column won't be filled.\n");
      return;
    };
    my $expectedRows = $col->rows();
    my $pd;
    my $index = 0;
    my $nrows = 0;
    while(<$fh>) {	
      chomp;
      next if /^\s*[\#!]/; # skip comment lines
      next if /^\s*$/; # skip blank lines
      my $lp = LineParser->new($_);
      my $c = $lp->getField($colnum);

      print STDERR "Parsing: ",$_,"\n" if($ENV{"VERBOSE"});

      ##
      ## STRING column
      ##
      if ($col->type() == $DAL::STRING) {
	# get the data vector only on the first pass
	$pd = $col->data() if($nrows == 0);
        ##
        ## Quoted string
        ##
        if($c->isString()) {
          if($c->elements() > $col->elements()) {
            warn("Truncating string at row $nrows, column ",$col->name(),"\n") if($ENV{"VERBOSE"});
          } elsif ($c->elements() < $col->elements()) {
            warn("Padding string with blanks at row $nrows, column ",$col->name(),"\n")
	      if($ENV{"VERBOSE"});
          }
          for(my $i = 0; $i <= min($col->elements(), $c->elements()); $i++) {
            $pd->set($c->getField($i), $nrows * $col->elements() + $i);
            for(my $i = min($col->elements(), $c->elements());
                $i < $col->elements(); $i++) {
              $pd->set(" ", $nrows * $col->elements() + $i);
            }
          }
        } else {
          ##
          ## Bare string
          ##
          # not a quoted string: convert to string only if
          # there is one item
          croak("Wrong format at row $nrows of $fname, column $colnum.\n")
	    if($c->elements() != 1);
          my $form = sprintf "%%-%ds", $col->elements();
          my $this = sprintf $form, $c->getField(0);              
          for(my $i = 0; $i < $col->elements(); $i++) {                
            $pd->set(substr($this, $i, 1), $nrows * $col->elements() + $i);
          }
        }	    
      } else {            
	##
	## Numerical columns
	##
	## Variable length vector columns
	##
	if($col->isVariableLength()){
	  print $col->name()," is variable length\n" if($ENV{"VERBOSE"});
	  $col->setCellSize(row => $nrows, size => $c->elements());
	  # here we get the data vector one row at a time
	  $pd = $col->cellData(row => $nrows);
	  for(my $i = 0; $i < $c->elements(); $i++) {
	    $pd->set($c->getField($i), $i);
	  }	  
	} else {
	  # get the data vector only on the first pass
	  $pd = $col->data() if($nrows == 0);
	  if($c->elements() < $col->elements() and not $padding) {
	    croak(sprintf "Need %d value(s) but found %d at row %s, and padding is not enabled.\n", $col->elements(), $c->elements(), $nrows);
	  } elsif($c->elements() > $col->elements() and not $padding) {
	    croak(sprintf "Need %d value(s) but found %d at row %s, check your data file.\n", $col->elements(), $c->elements(), $nrows);
	  }
	  my ($min, $max) = (sort {$a <=> $b} ($col->elements(), $c->elements()));	
	  for(my $i = 0; $i < $min; $i++) {
	    $pd->set($c->getField($i), $nrows * $col->elements() + $i);
	  }
	  for(my $i = $min; $i < $col->elements(); ++$i){
	    $pd->set(0, $nrows * $col->elements() + $i);
	  }
	}
      }
      $nrows++;
      undef $lp;

      if($nrows > $expectedRows){
	croak("There are more rows in the data file than I expected to create in the dataset. Check what files you use in rows and filleColumn.");
      }
      
    }
    undef $fh;
    $col->release();
    printf STDERR "%d rows\n", $nrows if($ENV{"VERBOSE"});
  }

#
# fillTable(<ptr>, <filename>)
#
sub fillTable
  {
    print "fillTable not implemented";
  }


#
# fillArray(<prt>, <filename>)
#
#  fills a DAL Array using data from an ASCII file. It expects
#     the image to be unwrapped (C style array indexing)
#
#   + <ptr>: pointer to the array to be filled (variable arr from deceit)
#   + <filename>: file containing the ASCII data to be read into the array
#
sub fillArray
  {
    if ((@_ != 2)) {
      croak("Usage: fillArray(arr, \$fname)\n");
    }
    my ($ptr, $fname) = @_;
    # this is a hack: take the address of the array (passed from
    # deceit), cast it into a pointer to an arry, then stuff it into an
    # array object.
    my $p = makeArrayPtr($ptr);
    my $array = DAL::array->_new;
    $array->ptr($p);
    printf "\nFilling array %s from %s -->", $array->name(), $fname if ($ENV{"VERBOSE"});
    my $fh = FileHandle->new();
    $fh->open($fname, "r") || do {
      carp("$fname does not exist. The array will be left empty.\n");
      return;
    };
    my $expectedElements = $array->elements();
    my $pd = $array->data();
    my $i=0;
    while(<$fh>) {
      next if /^\s*[\#!]/; # skip comment lines
      next if /^\s*$/; # skip blank lines
      chomp;
      foreach (split) {
        $pd->set($_,$i);      
        if (++$i > $expectedElements) {
          carp("Will not fill more then $expectedElements array cells.\n");
          last;
        }
      }
    }
    undef $fh;
  }

# This only works for 2D arrays
sub fillArrayFromTuple {
    if (@_ != 2) {
	croak("Usage: fillArrayFromTuple(arr, \$fname)\n");
    }
    my ($ptr, $fname) = @_;
    my $p = makeArrayPtr($ptr);
    my $array = DAL::array->_new;
    $array->ptr($p);
    my $arraydata = $array->data();

    my @dims = $array->dimensions();

    my $fh = FileHandle->new("$fname", "r") or die ("Cannot open $fname\n");
    while(<$fh>){
	next if /^\s*[#!]/; # skip comment lines
	next if /^\s*$/; # skip blank lines
	chomp;
	my @line = split;
	my $index = $line[0] * $dims[1] + $line[1];
	$arraydata->set($line[2],$index);
    }
}

#
# fillArrayFromFits(<prt>, <fitsname>[, <extname>])
#
#  fills a DAL Array using data from a FITS file.
#
#   + <ptr>: pointer to the array to be filled (variable arr from deceit)
#   + <fitsname>: file containing the FITS data to be read into the
#     array
#   + <extname>: optional name of the extension from which to extract
#     the data. If this is omitted the name is PRIMARY.
#
sub fillArrayFromFits {
    if (@_ < 2 or @_ > 3) {
	croak("Usage: fillArrayFromFits(arr, \$fitsname[, \$extname])\n");
    }
    my ($ptr, $fits, $extname) = @_;
    $extname = "PRIMARY" unless defined $extname;
    my $p = makeArrayPtr($ptr);
    my $array = DAL::array->_new;
    $array->ptr($p);

    my $ds = DAL::dataset->new(name => $fits, accessmode => $DAL::READ);
    my $fromarray = $ds->array(name => $extname);
    my $fromdata = $fromarray->data();
    my $arraydata = $array->data();

    for(my $i=0; $i<$fromarray->elements(); $i++){
	$arraydata->set($fromdata->value($i),$i);
    }
    $ds->release();
}

#
# fillConstArray(<ptr>, <value>) 
#
#  fills a DAL Array with a constant value
#  
sub fillConstArray {
  if (@_ != 2) {
    croak("Usage: fillConstArray(arr, \$value)\n");
  }
  my ($ptr, $c) = @_;
  my $p = makeArrayPtr($ptr);
  my $array = DAL::array->_new;
  $array->ptr($p);
  my $pd = $array->data();
  print "Filling the array ",$array->name(), " with ", $array->elements(), " times $c\n" if ($ENV{"VERBOSE"});
  for(my $i = 0; $i < $array->elements(); $i++){
    $pd->set($c,$i);
  }
  
}

#
# fillConstColumn(<ptr>, <value>)
#
#  fills a DAL column with a constant value
#
sub fillConstColumn {
  if (@_ != 2) {
    croak("Usage: fillConstColumn(col, \$value)\n");
  }
  my ($ptr, $c) = @_;
  my $p = makeColumnPtr($ptr);
  my $col = DAL::column->_new();
  $col->ptr($p);
  print "Filling the column ", $col->name()," with $c\n" if ($ENV{"VERBOSE"});
  if ($col->type() == $DAL::STRING) {
    for ( my $i = 0; $i < $col->rows(); $i++) {
      setStringCell($col, $i, $c);
    }
  } else {
    my $cd = $col->data();
    for (my $i =0; $i < $col->elements() * $col->rows(); $i++) {
      $cd->set($c, $i);
    }
  }
  return;
}

#
# fillAttribute(<ptr>, <filename>[, <string>])
#
# fills a DAL attribute with values from and ASCII file
#
#  + <ptr>: pointer to the attribute to be filled (variable att from deceit)
#  + <filename>: ascii data file containing (name, value) pairs. By
#    default "name" is the name of the attribute, unless the optional
#    <string> is specified. (the latter is not implemented yet).
#
sub fillAttribute {
    if (@_ != 2) {
	croak("Usage: fillAttribute(att, \$filename)\n");
    }
    my ($ptr, $fname) = @_;
    my $p = makeAttributePtr($ptr);
    my $att = DAL::attribute->_new();
    $att->ptr($p);
    my $fh = FileHandle->new($fname, "r") or do {
	carp("$fname does not exist. The attribute won't be filled.\n");
	return;
    };
    while(<$fh>){
	next if /^#/;
	next if /^\s*$/;
	chomp;
	my($name,$value) = split;
	print "**$name**$value**\n";
	if($name eq $att->name()){
	    $att->set($value);
	    last;
	}
    }
    undef $fh;
}

sub callUserFunction {
    my ($ptr, $func) = @_;
    require "$func.pl";
    eval "$func($ptr)";
}
